from typing import Any, List


def reserve(list: List[Any], capacity: int) -> None:
    """Reserve list capacity."""


def capacity(list: List[Any]) -> int:
    """Return list capacity."""


def allocated_bytes(list: List[Any]) -> int:
    """Return list allocated memory size."""


def shrink_to_fit(list: List[Any]) -> None:
    """Shrink to fit list capacity."""
